@@ -21,6 +21,8 @@ from eth2spec.utils.hash_function import hash
 
 SSZObject = TypeVar('SSZObject', bound=View)
 
+CONFIG_NAME = 'mainnet'
+
 
 fork = 'phase0'
 
@@ -73,8 +75,10 @@ class BLSSignature(Bytes96):
     pass
 
 
-def ceillog2(x: uint64) -> int:
-    return (x - 1).bit_length()
+def ceillog2(x: int) -> uint64:
+    if x < 1:
+        raise ValueError(f"ceillog2 accepts only positive values, x={x}")
+    return uint64((x - 1).bit_length())
 
 
 GENESIS_SLOT = Slot(0)
@@ -84,7 +88,7 @@ BASE_REWARDS_PER_EPOCH = uint64(4)
 DEPOSIT_CONTRACT_TREE_DEPTH = uint64(2**5)
 JUSTIFICATION_BITS_LENGTH = uint64(4)
 ENDIANNESS = 'little'
-ETH1_FOLLOW_DISTANCE = uint64(2**10)
+ETH1_FOLLOW_DISTANCE = uint64(2**11)
 MAX_COMMITTEES_PER_SLOT = uint64(2**6)
 TARGET_COMMITTEE_SIZE = uint64(2**7)
 MAX_VALIDATORS_PER_COMMITTEE = uint64(2**11)
@@ -92,7 +96,7 @@ MIN_PER_EPOCH_CHURN_LIMIT = uint64(2**2)
 CHURN_LIMIT_QUOTIENT = uint64(2**16)
 SHUFFLE_ROUND_COUNT = uint64(90)
 MIN_GENESIS_ACTIVE_VALIDATOR_COUNT = uint64(2**14)
-MIN_GENESIS_TIME = uint64(1578009600)
+MIN_GENESIS_TIME = uint64(1606824000)
 HYSTERESIS_QUOTIENT = uint64(4)
 HYSTERESIS_DOWNWARD_MULTIPLIER = uint64(1)
 HYSTERESIS_UPWARD_MULTIPLIER = uint64(5)
@@ -102,7 +106,7 @@ EJECTION_BALANCE = Gwei(2**4 * 10**9)
 EFFECTIVE_BALANCE_INCREMENT = Gwei(2**0 * 10**9)
 GENESIS_FORK_VERSION = Version('0x00000000')
 BLS_WITHDRAWAL_PREFIX = Bytes1('0x00')
-GENESIS_DELAY = uint64(172800)
+GENESIS_DELAY = uint64(604800)
 SECONDS_PER_SLOT = uint64(12)
 SECONDS_PER_ETH1_BLOCK = uint64(14)
 MIN_ATTESTATION_INCLUSION_DELAY = uint64(2**0)
@@ -110,7 +114,7 @@ SLOTS_PER_EPOCH = uint64(2**5)
 MIN_SEED_LOOKAHEAD = uint64(2**0)
 MAX_SEED_LOOKAHEAD = uint64(2**2)
 MIN_EPOCHS_TO_INACTIVITY_PENALTY = uint64(2**2)
-EPOCHS_PER_ETH1_VOTING_PERIOD = uint64(2**5)
+EPOCHS_PER_ETH1_VOTING_PERIOD = uint64(2**6)
 SLOTS_PER_HISTORICAL_ROOT = uint64(2**13)
 MIN_VALIDATOR_WITHDRAWABILITY_DELAY = uint64(2**8)
 SHARD_COMMITTEE_PERIOD = uint64(2**8)
@@ -121,8 +125,9 @@ VALIDATOR_REGISTRY_LIMIT = uint64(2**40)
 BASE_REWARD_FACTOR = uint64(2**6)
 WHISTLEBLOWER_REWARD_QUOTIENT = uint64(2**9)
 PROPOSER_REWARD_QUOTIENT = uint64(2**3)
-INACTIVITY_PENALTY_QUOTIENT = uint64(2**24)
-MIN_SLASHING_PENALTY_QUOTIENT = uint64(2**5)
+INACTIVITY_PENALTY_QUOTIENT = uint64(2**26)
+MIN_SLASHING_PENALTY_QUOTIENT = uint64(2**7)
+PROPORTIONAL_SLASHING_MULTIPLIER = uint64(1)
 MAX_PROPOSER_SLASHINGS = 2**4
 MAX_ATTESTER_SLASHINGS = 2**1
 MAX_ATTESTATIONS = 2**7
@@ -140,6 +145,7 @@ TARGET_AGGREGATORS_PER_COMMITTEE = 2**4
 RANDOM_SUBNETS_PER_VALIDATOR = 2**0
 EPOCHS_PER_RANDOM_SUBNET_SUBSCRIPTION = 2**8
 ATTESTATION_SUBNET_COUNT = 64
+SAFETY_DECAY = uint64(10)
 
 
 apply_constants_config(globals())
@@ -344,6 +350,25 @@ class SignedAggregateAndProof(Container):
     signature: BLSSignature
 
 
+@dataclass(eq=True, frozen=True)
+class LatestMessage(object):
+    epoch: Epoch
+    root: Root
+
+
+@dataclass
+class Store(object):
+    time: uint64
+    genesis_time: uint64
+    justified_checkpoint: Checkpoint
+    finalized_checkpoint: Checkpoint
+    best_justified_checkpoint: Checkpoint
+    blocks: Dict[Root, BeaconBlock] = field(default_factory=dict)
+    block_states: Dict[Root, BeaconState] = field(default_factory=dict)
+    checkpoint_states: Dict[Checkpoint, BeaconState] = field(default_factory=dict)
+    latest_messages: Dict[ValidatorIndex, LatestMessage] = field(default_factory=dict)
+
+
 def integer_squareroot(n: uint64) -> uint64:
     """
     Return the largest integer ``x`` such that ``x**2 <= n``.
@@ -495,7 +520,7 @@ def compute_committee(indices: Sequence[ValidatorIndex],
     Return the committee corresponding to ``indices``, ``seed``, ``index``, and committee ``count``.
     """
     start = (len(indices) * index) // count
-    end = (len(indices) * (index + 1)) // count
+    end = (len(indices) * uint64(index + 1)) // count
     return [indices[compute_shuffled_index(uint64(i), uint64(len(indices)), seed)] for i in range(start, end)]
 
 
@@ -809,7 +834,7 @@ def is_valid_genesis_state(state: BeaconState) -> bool:
     return True
 
 
-def state_transition(state: BeaconState, signed_block: SignedBeaconBlock, validate_result: bool=True) -> BeaconState:
+def state_transition(state: BeaconState, signed_block: SignedBeaconBlock, validate_result: bool=True) -> None:
     block = signed_block.message
     # Process slots (including those with no blocks) since block
     process_slots(state, block.slot)
@@ -821,8 +846,6 @@ def state_transition(state: BeaconState, signed_block: SignedBeaconBlock, valida
     # Verify state root
     if validate_result:
         assert block.state_root == hash_tree_root(state)
-    # Return post-state
-    return state
 
 
 def verify_block_signature(state: BeaconState, signed_block: SignedBeaconBlock) -> bool:
@@ -897,6 +920,8 @@ def get_attesting_balance(state: BeaconState, attestations: Sequence[PendingAtte
 
 
 def process_justification_and_finalization(state: BeaconState) -> None:
+    # Initial FFG checkpoint values have a `0x00` stub for `root`.
+    # Skip FFG updates in the first two epochs to avoid corner cases that might result in modifying this stub.
     if get_current_epoch(state) <= GENESIS_EPOCH + 1:
         return
 
@@ -1024,7 +1049,7 @@ def get_inclusion_delay_deltas(state: BeaconState) -> Tuple[Sequence[Gwei], Sequ
             if index in get_attesting_indices(state, a.data, a.aggregation_bits)
         ], key=lambda a: a.inclusion_delay)
         rewards[attestation.proposer_index] += get_proposer_reward(state, index)
-        max_attester_reward = get_base_reward(state, index) - get_proposer_reward(state, index)
+        max_attester_reward = Gwei(get_base_reward(state, index) - get_proposer_reward(state, index))
         rewards[index] += Gwei(max_attester_reward // attestation.inclusion_delay)
 
     # No penalties associated with inclusion delay
@@ -1077,6 +1102,7 @@ def get_attestation_deltas(state: BeaconState) -> Tuple[Sequence[Gwei], Sequence
 
 
 def process_rewards_and_penalties(state: BeaconState) -> None:
+    # No rewards are applied at the end of `GENESIS_EPOCH` because rewards are for work done in the previous epoch
     if get_current_epoch(state) == GENESIS_EPOCH:
         return
 
@@ -1110,10 +1136,11 @@ def process_registry_updates(state: BeaconState) -> None:
 def process_slashings(state: BeaconState) -> None:
     epoch = get_current_epoch(state)
     total_balance = get_total_active_balance(state)
+    adjusted_total_slashing_balance = min(sum(state.slashings) * PROPORTIONAL_SLASHING_MULTIPLIER, total_balance)
     for index, validator in enumerate(state.validators):
         if validator.slashed and epoch + EPOCHS_PER_SLASHINGS_VECTOR // 2 == validator.withdrawable_epoch:
             increment = EFFECTIVE_BALANCE_INCREMENT  # Factored out from penalty numerator to avoid uint64 overflow
-            penalty_numerator = validator.effective_balance // increment * min(sum(state.slashings) * 3, total_balance)
+            penalty_numerator = validator.effective_balance // increment * adjusted_total_slashing_balance
             penalty = penalty_numerator // total_balance * increment
             decrease_balance(state, ValidatorIndex(index), penalty)
 
@@ -1127,7 +1154,7 @@ def process_final_updates(state: BeaconState) -> None:
     # Update effective balances with hysteresis
     for index, validator in enumerate(state.validators):
         balance = state.balances[index]
-        HYSTERESIS_INCREMENT = EFFECTIVE_BALANCE_INCREMENT // HYSTERESIS_QUOTIENT
+        HYSTERESIS_INCREMENT = uint64(EFFECTIVE_BALANCE_INCREMENT // HYSTERESIS_QUOTIENT)
         DOWNWARD_THRESHOLD = HYSTERESIS_INCREMENT * HYSTERESIS_DOWNWARD_MULTIPLIER
         UPWARD_THRESHOLD = HYSTERESIS_INCREMENT * HYSTERESIS_UPWARD_MULTIPLIER
         if (
@@ -1347,30 +1374,9 @@ def process_voluntary_exit(state: BeaconState, signed_voluntary_exit: SignedVolu
     initiate_validator_exit(state, voluntary_exit.validator_index)
 
 
-@dataclass(eq=True, frozen=True)
-class LatestMessage(object):
-    epoch: Epoch
-    root: Root
-
-
-@dataclass
-class Store(object):
-    time: uint64
-    genesis_time: uint64
-    justified_checkpoint: Checkpoint
-    finalized_checkpoint: Checkpoint
-    best_justified_checkpoint: Checkpoint
-    blocks: Dict[Root, BeaconBlock] = field(default_factory=dict)
-    block_states: Dict[Root, BeaconState] = field(default_factory=dict)
-    checkpoint_states: Dict[Checkpoint, BeaconState] = field(default_factory=dict)
-    latest_messages: Dict[ValidatorIndex, LatestMessage] = field(default_factory=dict)
-
-
-def get_forkchoice_store(anchor_state: BeaconState) -> Store:
-    anchor_block_header = copy(anchor_state.latest_block_header)
-    if anchor_block_header.state_root == Bytes32():
-        anchor_block_header.state_root = hash_tree_root(anchor_state)
-    anchor_root = hash_tree_root(anchor_block_header)
+def get_forkchoice_store(anchor_state: BeaconState, anchor_block: BeaconBlock) -> Store:
+    assert anchor_block.state_root == hash_tree_root(anchor_state)
+    anchor_root = hash_tree_root(anchor_block)
     anchor_epoch = get_current_epoch(anchor_state)
     justified_checkpoint = Checkpoint(epoch=anchor_epoch, root=anchor_root)
     finalized_checkpoint = Checkpoint(epoch=anchor_epoch, root=anchor_root)
@@ -1380,7 +1386,7 @@ def get_forkchoice_store(anchor_state: BeaconState) -> Store:
         justified_checkpoint=justified_checkpoint,
         finalized_checkpoint=finalized_checkpoint,
         best_justified_checkpoint=justified_checkpoint,
-        blocks={anchor_root: anchor_block_header},
+        blocks={anchor_root: copy(anchor_block)},
         block_states={anchor_root: copy(anchor_state)},
         checkpoint_states={justified_checkpoint: copy(anchor_state)},
     )
@@ -1471,11 +1477,10 @@ def get_head(store: Store) -> Root:
     blocks = get_filtered_block_tree(store)
     # Execute the LMD-GHOST fork choice
     head = store.justified_checkpoint.root
-    justified_slot = compute_start_slot_at_epoch(store.justified_checkpoint.epoch)
     while True:
         children = [
             root for root in blocks.keys()
-            if blocks[root].parent_root == head and blocks[root].slot > justified_slot
+            if blocks[root].parent_root == head
         ]
         if len(children) == 0:
             return head
@@ -1577,7 +1582,8 @@ def on_block(store: Store, signed_block: SignedBeaconBlock) -> None:
     assert get_ancestor(store, block.parent_root, finalized_slot) == store.finalized_checkpoint.root
 
     # Check the block is valid and compute the post-state
-    state = state_transition(pre_state, signed_block, True)
+    state = pre_state.copy()
+    state_transition(state, signed_block, True)
     # Add new block to the store
     store.blocks[hash_tree_root(block)] = block
     # Add new state for this block to the store
@@ -1644,7 +1650,7 @@ def get_committee_assignment(state: BeaconState,
         * ``assignment[2]`` is the slot at which the committee is assigned
     Return None if no assignment.
     """
-    next_epoch = get_current_epoch(state) + 1
+    next_epoch = Epoch(get_current_epoch(state) + 1)
     assert epoch <= next_epoch
 
     start_slot = compute_start_slot_at_epoch(epoch)
@@ -1711,7 +1717,7 @@ def get_eth1_vote(state: BeaconState, eth1_chain: Sequence[Eth1Block]) -> Eth1Da
 def compute_new_state_root(state: BeaconState, block: BeaconBlock) -> Root:
     temp_state: BeaconState = state.copy()
     signed_block = SignedBeaconBlock(message=block)
-    temp_state = state_transition(temp_state, signed_block, validate_result=False)
+    state_transition(temp_state, signed_block, validate_result=False)
     return hash_tree_root(temp_state)
 
 
@@ -1732,7 +1738,7 @@ def compute_subnet_for_attestation(committees_per_slot: uint64, slot: Slot, comm
     Compute the correct subnet for an attestation for Phase 0.
     Note, this mimics expected Phase 1 behavior where attestations will be mapped to their shard subnet.
     """
-    slots_since_epoch_start = slot % SLOTS_PER_EPOCH
+    slots_since_epoch_start = uint64(slot % SLOTS_PER_EPOCH)
     committees_since_epoch_start = committees_per_slot * slots_since_epoch_start
 
     return uint64((committees_since_epoch_start + committee_index) % ATTESTATION_SUBNET_COUNT)
@@ -1775,6 +1781,27 @@ def get_aggregate_and_proof_signature(state: BeaconState,
     return bls.Sign(privkey, signing_root)
 
 
+def compute_weak_subjectivity_period(state: BeaconState) -> uint64:
+    weak_subjectivity_period = MIN_VALIDATOR_WITHDRAWABILITY_DELAY
+    validator_count = len(get_active_validator_indices(state, get_current_epoch(state)))
+    if validator_count >= MIN_PER_EPOCH_CHURN_LIMIT * CHURN_LIMIT_QUOTIENT:
+        weak_subjectivity_period += SAFETY_DECAY * CHURN_LIMIT_QUOTIENT // (2 * 100)
+    else:
+        weak_subjectivity_period += SAFETY_DECAY * validator_count // (2 * 100 * MIN_PER_EPOCH_CHURN_LIMIT)
+    return weak_subjectivity_period
+
+
+def is_within_weak_subjectivity_period(store: Store, ws_state: BeaconState, ws_checkpoint: Checkpoint) -> bool:
+    # Clients may choose to validate the input state against the input Weak Subjectivity Checkpoint
+    assert ws_state.latest_block_header.state_root == ws_checkpoint.root
+    assert compute_epoch_at_slot(ws_state.slot) == ws_checkpoint.epoch
+
+    ws_period = compute_weak_subjectivity_period(ws_state)
+    ws_state_epoch = compute_epoch_at_slot(ws_state.slot)
+    current_epoch = compute_epoch_at_slot(get_current_slot(store))
+    return current_epoch <= ws_state_epoch + ws_period
+
+
 def get_eth1_data(block: Eth1Block) -> Eth1Data:
     """
     A stub function return mocking Eth1Data.
